/*
 * This file is part of Locana.
 * 
 * Copyright (c) 2007 Thomas King <king@informatik.uni-mannheim.de>,
 * University of Mannheim, Germany
 * 
 * All rights reserved.
 *
 * Loclib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Loclib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Loclib; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package org.pi4.locana.misc;

import java.util.ArrayList;

public class AccessPoint {
	private String macAddr;
	private ArrayList<Integer> readings;
	private ArrayList<Integer> directions;
	private int histogram[];
	private int histogramCopy[];
	private double filteredHistogram[];
	private double dirMeans[];
	private double dirDevs[];
	private long counts[];
	private double gaussianMeans[];
	private double gaussianStds[];
	private double gaussianScales[];
	private int pollCount;
	
	
	public AccessPoint(String macAddr) {
		this.macAddr = macAddr;
		readings = new ArrayList<Integer>();
		directions = new ArrayList<Integer>();
		pollCount = 0;
	}
	
	public void incPollCount() {
		pollCount++;
	}
	
	public int getPollCount() {
		return pollCount;
	}
	
	public String getMacAddr() {
		return macAddr;
	}
	
	public void addReading(int reading, double direction) {
		int dir = new Long(Math.round(direction / 5.0)).intValue() * 5;
		if (dir == 360)
			dir = 0;
		readings.add(new Integer(reading));
		directions.add(dir);
	}
	
	public void createDirMeans() {
		double sums[] = new double[360];
		counts = new long[360];

		for (int i = 0; i < 360; i++) {
			sums[i] = 0;
			counts[i] = 0;
		}
		for (int i = 0; i < readings.size(); i++) {
			int rss = readings.get(i);
			if (rss <= 0) {
				int dir = directions.get(i);
				sums[dir] += rss;
				counts[dir]++;
			}
		}
		dirMeans = new double[360];
		for (int i = 0; i < 360; i++) {
			if (counts[i] == 0)
				dirMeans[i] = 999.0;
			else
				dirMeans[i] = sums[i] / counts[i];
		}
	}
	
	public double getDirMean(int direction) {
		return dirMeans[direction];
	}
	
	public double getDirDeviation(int direction) {
		return dirDevs[direction];
	}
	
	public long getDirCounts(int direction) {
		return counts[direction];
	}
	
	public double getMean(int start, int end) {
		double sum = 0.0;
		int count=0;
		for (int i = start; i <= end; i++) {
			sum += histogramCopy[i] * (i - 102);
			count += histogramCopy[i];
		}
		return sum / count;
	}
	
	public double getVariance(int lag) {
		double mean = getMean(0, 102);
		// Liste der readings ohne Lücken erstellen
		ArrayList<Integer> noGapList = new ArrayList<Integer>();
		for (int i = 0; i < readings.size(); i++) {
			int rss = 0;
			if ((rss = readings.get(i)) <= 0)
				noGapList.add(rss);
		}
		// Varianz dieser Liste bestimmen
		double sum = 0.0;
		int count = 0;
		for (int i = 0; i < (noGapList.size()) - lag; i++) {
			sum += (noGapList.get(i) - mean) * (noGapList.get(i + lag) - mean);
			count++;
		}
		return sum / count;
	}
	
	public void createDirDeviations() {
		double sums[] = new double[360];
		createDirMeans();
		for (int i = 0; i < (readings.size()); i++) {
			int rss = readings.get(i);
			int dir = directions.get(i);
			if (rss <= 0) {
				sums[dir] += ((rss - dirMeans[dir]) * (rss - dirMeans[dir]));
			}
		}
		dirDevs = new double[360];
		for (int i = 0; i < 360; i++) {
			if (counts[i] <= 1)
				dirDevs[i] = 0.0;
			else
				dirDevs[i] = Math.sqrt(sums[i] / (counts[i]-1));
		}
	}
	
	public double getStandardDeviation(int start, int end) {
		double mean = getMean(start, end);
		
		double sum = 0.0;
		int count = 0;
		for (int i = start; i <= end; i++) {
			sum += ((i - 102) - mean) * ((i - 102) - mean) * histogramCopy[i];
			count += histogramCopy[i];
		}
		if (count == 1)
			return 0.0;
		else
			return Math.sqrt(sum / (count - 1));
	}
	
	public void createHistogram() {
		// Histogramm-Arrays initialisieren
		histogram = new int[103];
		histogramCopy = new int[103];
		for (int i = 0; i < histogram.length; i++)
			histogram[i] = 0;
		// Histogramm erstellen
		for (int i = 0; i < readings.size(); i++) {
			int rss = readings.get(i);
			if (rss <= 0)
				histogram[rss + 102]++;
		}
		// Histogramm-Kopie (für mgauss) erstellen
		for (int i = 0; i < histogram.length; i++) {
			histogramCopy[i] = histogram[i];
		}
	}
	
	public int getRssFrequency(int rss) {
		return histogram[rss + 102];
	}
	
	public double getFilteredRssFrequency(int rss) {
		return filteredHistogram[rss + 102];
	}
	
	public int getRss(int index) {
		if (index < readings.size())
			return readings.get(index);
		else
			return 999;
	}
	
	public int getRssCount(int start, int end) {
		if (histogramCopy == null)
			return readings.size();
		int sum = 0;
		for (int i = start; i <= end; i++) {
			sum += histogramCopy[i];
		}
		return sum;
	}
	
	public void applyGaussianFilter() {
		filteredHistogram = new double[103];
		filteredHistogram[0] = 0.0;
		filteredHistogram[102] = 0.0;
		for (int i = 1; i < (histogramCopy.length - 1); i++) {
			filteredHistogram[i] = histogramCopy[i];
		}
	}
	
	public int getFilteredMax() {
		int maxIndex = -1;
		double maxValue = -1.0;
		for (int i = 0; i < filteredHistogram.length; i++) {
			if (filteredHistogram[i] > maxValue) {
				maxValue = filteredHistogram[i];
				maxIndex = i;
			}
		}
		return maxIndex;
	}
	
	public int getLeftBoundary(int start) {
		double startValue = filteredHistogram[start];
		double lastValue = startValue;
		
		boolean foundSmallerValue = false;
		
		int leftBoundary = 0;
		for (int i = (start - 1); i >= 0; i--) {
			double currentValue = filteredHistogram[i];
			if (currentValue < startValue)
				foundSmallerValue = true;
			if (foundSmallerValue) {
				if (currentValue >= lastValue) {
					i++;
					leftBoundary = i;
					break;
				} else if (currentValue == 0.0) {
					leftBoundary = i;
					break;
				}
				leftBoundary = i;
			}
			leftBoundary = i;
			lastValue = currentValue;
		}
		return leftBoundary;
	}

	public int getRightBoundary(int start) {
		double startValue = filteredHistogram[start];
		double lastValue = startValue;
		
		boolean foundSmallerValue = false;
		
		int rightBoundary = 0;
		for (int i = (start + 1); i < filteredHistogram.length; i++) {
			double currentValue = filteredHistogram[i];
			if (currentValue < startValue)
				foundSmallerValue = true;
			if (foundSmallerValue) {
				if (currentValue >= lastValue) {
					i--;
					rightBoundary = i;
					break;
				} else if (currentValue == 0.0) {
					rightBoundary = i;
					break;
				}
				rightBoundary = i;
			}
			rightBoundary = i;
			lastValue = currentValue;
		}
		return rightBoundary;
	}
	
	public void deleteFromHistograms(int start, int end) {
		for (int i = start; i <= end; i++) {
			histogramCopy[i] = 0;
			filteredHistogram[i] = 0.0;
		}
	}
	
	public void substractFromHistogramCopy(double mean, double std, double scale) {
		
		for (int i = 1; i <= 101; i++) {
			double p1 = approximationCumulativeDistributionFunction(new Double(i - 102.5), mean, std);
			double p2 = approximationCumulativeDistributionFunction(new Double(i - 101.5), mean, std);
			int newValue = new Long(Math.round(new Double(histogramCopy[i]) - (scale * (p2 - p1)))).intValue();
			histogramCopy[i] = (newValue < 0 ? 0 : newValue);
		}
	}
	
	public void normalize() {
		double normalizer = 0.0;
		int i;
		for (i = -102; i <= 0; i++)
			normalizer += getGaussianApproximationValue(i);
		System.out.println("normalizer is "+normalizer);
		for (i = 0; i < gaussianScales.length; i++) {
			gaussianScales[i] /= normalizer;
			System.out.println("scale[" + i + "] is now " + gaussianScales[i]);
		}
	}
	
	//FIXME is this an exact one-to-one copy of the loceva method?
	private double approximationCumulativeDistributionFunction(double x_, double avg, double stddev) {
		// special case: stddev == 0.0
		if (stddev == 0.0) {
			if (x_ >= avg) return 1;
			return 0;
		}
		
		double x = (x_ - avg) / stddev;
		if (((x_ - avg) / stddev) < 0.0) x *= -1;
		// Z(x)
		double z = 1.0 / ((2.49089 + (1.466003 * x * x) + (-0.024393 * x * x * x * x + (0.178257 * x * x * x * x * x * x))));	
		
		// P(x)
		double t = 1.0 / (1.0 + (0.33267 * x));
		double approx = 1.0 - (z * ((0.4361836 * t) + (-0.1201676 * t * t) + (0.9372980 * t * t * t)));
		if (((x_ - avg) / stddev) < 0.0) approx = 1 - approx;
		
		return approx;
	}

	public double getGaussianApproximationValue(int rss) {
		double approximatedValue = 0.0;
		double a = 0.0;
		for (int j = 0; j < gaussianMeans.length; j++) {
			if (gaussianScales[j] == 0.0)
				a = 0.0;
			else {
				a = approximationCumulativeDistributionFunction(new Double(rss + 0.5), gaussianMeans[j], gaussianStds[j]);
				a -= approximationCumulativeDistributionFunction(new Double(rss - 0.5), gaussianMeans[j], gaussianStds[j]);
				a *= gaussianScales[j];
			}
			approximatedValue += a;
		}
		return approximatedValue;
	}
	
	public void setGaussianApproximation(double means[], double stds[], double scales[]) {
		gaussianMeans = new double[means.length];
		for (int i = 0; i < means.length; i++)
			gaussianMeans[i] = means[i];
		gaussianStds = new double[stds.length];
		for (int i = 0; i < stds.length; i++)
			gaussianStds[i] = stds[i];
		gaussianScales = new double[scales.length];
		for (int i = 0; i < scales.length; i++)
			gaussianScales[i] = scales[i];
	}
}
