/*
 * This file is part of Locana.
 * 
 * Copyright (c) 2007 Thomas King <king@informatik.uni-mannheim.de>,
 * University of Mannheim, Germany
 * 
 * All rights reserved.
 *
 * Loclib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Loclib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Loclib; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package org.pi4.locana.radiomap;

import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.jibble.epsgraphics.EpsGraphics2D;
import org.pi4.locutil1.MACAddress;
import org.pi4.locutil2.GeoPosition;
import org.pi4.locutil2.Statistics;

@SuppressWarnings("serial")
public class RadioMap extends JFrame {
	private static BufferedReader offlineFile;
	private static BufferedReader onlineFile;
	private Map<MACAddress, Map<GeoPosition, List<Double>>> offlineMap = new HashMap<MACAddress, Map<GeoPosition, List<Double>>>();
	private Map<MACAddress, Map<GeoPosition, Double>> offlineMapOfMeans   = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	private Map<MACAddress, Map<GeoPosition, Double>> offlineMapOfStdDevs = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	private Map<MACAddress, Map<GeoPosition, Double>> offlineMapOfMaxs    = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	private Map<MACAddress, Map<GeoPosition, Double>> offlineMapOfMins    = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	private Map<MACAddress, Map<GeoPosition, Double>> offlineMapOfAmounts = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	
	private Map<MACAddress, Map<GeoPosition, List<Double>>> onlineMap = new HashMap<MACAddress, Map<GeoPosition, List<Double>>>();
	private Map<MACAddress, Map<GeoPosition, Double>> onlineMapOfMeans   = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	private Map<MACAddress, Map<GeoPosition, Double>> onlineMapOfStdDevs = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	private Map<MACAddress, Map<GeoPosition, Double>> onlineMapOfMaxs    = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	private Map<MACAddress, Map<GeoPosition, Double>> onlineMapOfMins    = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	private Map<MACAddress, Map<GeoPosition, Double>> onlineMapOfAmounts = new HashMap<MACAddress, Map<GeoPosition, Double>>();
	
	private Map<GeoPositionStep, Map<GeoPosition, Double>> offlineRiceProbabilities = new HashMap<GeoPositionStep, Map<GeoPosition, Double>>();
	
	private ArrayList<GeoPosition> realPositionList = new ArrayList<GeoPosition>();
	private ArrayList<Double> orientationList = new ArrayList<Double>();
	private GridBagLayout gbLayout = new GridBagLayout();
	private GridBagConstraints gbConstraints = new GridBagConstraints();
	private JComboBox apChoice, realPosChoice, orientationChoice, stepChoice, scaleChoice, gridChoice, gradientChoice, measureChoice, startPosChoice, apPosChoice;
    private JComboBox markersChoice, tilesChoice;
    private JCheckBox valuesCheckBox;
    private JLabel infoText;
	private RadioMapViewport viewport;
	private double avgGlobalMinValue = Double.MAX_VALUE;
	private double avgGlobalMaxValue = Double.MIN_VALUE;
	private double stdDevGlobalMinValue = Double.MAX_VALUE;
	private double stdDevGlobalMaxValue = Double.MIN_VALUE;
	private double maxGlobalMinValue = Double.MAX_VALUE;
	private double maxGlobalMaxValue = Double.MIN_VALUE;
	private double minGlobalMinValue = Double.MAX_VALUE;
	private double minGlobalMaxValue = Double.MIN_VALUE;
	private double amountGlobalMinValue = Double.MAX_VALUE;
	private double amountGlobalMaxValue = Double.MIN_VALUE;
	private Map<MACAddress, GeoPosition> apPosition = new HashMap<MACAddress, GeoPosition>();
	private static double minGridSpacing = 0.5;
    private static double maxGridSpacing = 5.0;
    private static double gridSpacingSteps = 0.5;
	private static String mapMode;
	
	public RadioMap() {
		int steps = 0;
		
		try {
			if (mapMode.equals("tracefile")) {
				avgGlobalMinValue = 0.0;
				avgGlobalMaxValue = -103.0;
				
				if (offlineFile != null) TraceAnalyzer.analyzeTracefile(offlineFile, offlineMap, apPosition, orientationList);
				if (onlineFile != null) TraceAnalyzer.analyzeTracefile(onlineFile,  onlineMap, apPosition, orientationList);
			} else if (mapMode.equals("probability")) {
				/*
				 * Format:
				 * real=x,y,z,o
				 * step=i;x1,y1,p1;x2,y2,p2;....
				 */
				
				avgGlobalMinValue = 1.0;
				avgGlobalMaxValue = 0.0;
				
				double[] minmax = new double[2];
				steps = TraceAnalyzer.analyzeRiceLog(offlineFile, offlineRiceProbabilities, realPositionList, minmax);
				avgGlobalMinValue = minmax[0];
				avgGlobalMaxValue = minmax[1];
				
			} else System.err.println("Unknown map mode");
			
			// close input file
			offlineFile.close();
		} catch (IOException ex) {
			System.out.println("Couldn't read from the input file!");
			ex.printStackTrace();
			System.exit(-1);
		}
		
		// create the viewport
        viewport = new RadioMapViewport();
        
		if (mapMode.equals("tracefile")) {
            viewport.setInfoMode(RadioMapViewport.INFOMODE_TRACE);
			viewport.setValueRange(-103, 0);
			viewport.setColorBarScaleFactor(5);
			viewport.setValueScaleFactor(1);
			// calculate the means and create the choice box for the aps
			apChoice = new JComboBox();
			apChoice.addItem(" ");
			
			Set<MACAddress> apOfflineSet = new HashSet<MACAddress>();
			Iterator<MACAddress> apIterator = offlineMap.keySet().iterator();
			while (apIterator.hasNext()) {
				MACAddress key = apIterator.next();
				apOfflineSet.add(key);
				Map<GeoPosition, List<Double>> mapOfListOfDouble = offlineMap.get(key);
				Iterator<GeoPosition> posIterator = mapOfListOfDouble.keySet().iterator();
								
				Map<GeoPosition, Double> posMean;
				if (offlineMapOfMeans.containsKey(key)) posMean = offlineMapOfMeans.get(key); else posMean = new HashMap<GeoPosition, Double>();
				
				Map<GeoPosition, Double> posStdDev;
				if (offlineMapOfStdDevs.containsKey(key)) posStdDev = offlineMapOfStdDevs.get(key); else posStdDev = new HashMap<GeoPosition, Double>();
				
				Map<GeoPosition, Double> posMax;
				if (offlineMapOfMaxs.containsKey(key)) posMax = offlineMapOfMaxs.get(key); else posMax = new HashMap<GeoPosition, Double>();
				
				Map<GeoPosition, Double> posMin;
				if (offlineMapOfMins.containsKey(key)) posMin = offlineMapOfMins.get(key); else posMin = new HashMap<GeoPosition, Double>();
				
				Map<GeoPosition, Double> posAmount;
				if (offlineMapOfMins.containsKey(key)) posAmount = offlineMapOfMins.get(key); else posAmount = new HashMap<GeoPosition, Double>();
				
				while (posIterator.hasNext()) {
					GeoPosition pos = posIterator.next();
					double avg = Statistics.avg(mapOfListOfDouble.get(pos));
					double stdDev = Statistics.stdDev(mapOfListOfDouble.get(pos));
					double max = Statistics.max(mapOfListOfDouble.get(pos));
					double min = Statistics.min(mapOfListOfDouble.get(pos));
					double amount = mapOfListOfDouble.get(pos).size();

					// update global min and max values
					if (avg < avgGlobalMinValue)
						avgGlobalMinValue = avg;
					if (avg > avgGlobalMaxValue)
						avgGlobalMaxValue = avg;
					if (stdDev < stdDevGlobalMinValue)
						stdDevGlobalMinValue = stdDev;
					if (stdDev > stdDevGlobalMaxValue)
						stdDevGlobalMaxValue = stdDev;
					if (max < maxGlobalMinValue)
						maxGlobalMinValue = max;
					if (max > maxGlobalMaxValue)
						maxGlobalMaxValue = max;
					if (min < minGlobalMinValue)
						minGlobalMinValue = min;
					if (min > minGlobalMaxValue)
						minGlobalMaxValue = min;
					if (amount < amountGlobalMinValue)
						amountGlobalMinValue = amount;
					if (amount > amountGlobalMaxValue)
						amountGlobalMaxValue = amount;
					
					posMean.put(pos, avg);
					posStdDev.put(pos, stdDev);
					posMax.put(pos, max);
					posMin.put(pos, min);
					posAmount.put(pos, amount);
				}
				offlineMapOfMeans.put(key, posMean);
				offlineMapOfStdDevs.put(key, posStdDev);
				offlineMapOfMaxs.put(key, posMax);
				offlineMapOfMins.put(key, posMin);
				offlineMapOfAmounts.put(key, posAmount);
			}
			
			Set<MACAddress> apOnlineSet = new HashSet<MACAddress>();
			apIterator = onlineMap.keySet().iterator();
			while (apIterator.hasNext()) {
				MACAddress key = apIterator.next();
				apOnlineSet.add(key);
				Map<GeoPosition, List<Double>> mapOfListOfDouble = onlineMap.get(key);
				Iterator<GeoPosition> posIterator = mapOfListOfDouble.keySet().iterator();
								
				Map<GeoPosition, Double> posMean;
				if (onlineMapOfMeans.containsKey(key)) posMean = onlineMapOfMeans.get(key); else posMean = new HashMap<GeoPosition, Double>();
				
				Map<GeoPosition, Double> posStdDev;
				if (onlineMapOfStdDevs.containsKey(key)) posStdDev = onlineMapOfStdDevs.get(key); else posStdDev = new HashMap<GeoPosition, Double>();
				
				Map<GeoPosition, Double> posMax;
				if (onlineMapOfMaxs.containsKey(key)) posMax = onlineMapOfMaxs.get(key); else posMax = new HashMap<GeoPosition, Double>();
				
				Map<GeoPosition, Double> posMin;
				if (onlineMapOfMins.containsKey(key)) posMin = onlineMapOfMins.get(key); else posMin = new HashMap<GeoPosition, Double>();
				
				Map<GeoPosition, Double> posAmount;
				if (onlineMapOfMins.containsKey(key)) posAmount = onlineMapOfMins.get(key); else posAmount = new HashMap<GeoPosition, Double>();
				
				while (posIterator.hasNext()) {
					GeoPosition pos = posIterator.next();
					double avg = Statistics.avg(mapOfListOfDouble.get(pos));
					double stdDev = Statistics.stdDev(mapOfListOfDouble.get(pos));
					double max = Statistics.max(mapOfListOfDouble.get(pos));
					double min = Statistics.min(mapOfListOfDouble.get(pos));
					double amount = mapOfListOfDouble.get(pos).size();

					// update global min and max values
					if (avg < avgGlobalMinValue)
						avgGlobalMinValue = avg;
					if (avg > avgGlobalMaxValue)
						avgGlobalMaxValue = avg;
					if (stdDev < stdDevGlobalMinValue)
						stdDevGlobalMinValue = stdDev;
					if (stdDev > stdDevGlobalMaxValue)
						stdDevGlobalMaxValue = stdDev;
					if (max < maxGlobalMinValue)
						maxGlobalMinValue = max;
					if (max > maxGlobalMaxValue)
						maxGlobalMaxValue = max;
					if (min < minGlobalMinValue)
						minGlobalMinValue = min;
					if (min > minGlobalMaxValue)
						minGlobalMaxValue = min;
					if (amount < amountGlobalMinValue)
						amountGlobalMinValue = amount;
					if (amount > amountGlobalMaxValue)
						amountGlobalMaxValue = amount;
					
					posMean.put(pos, avg);
					posStdDev.put(pos, stdDev);
					posMax.put(pos, max);
					posMin.put(pos, min);
					posAmount.put(pos, amount);
				}
				onlineMapOfMeans.put(key, posMean);
				onlineMapOfStdDevs.put(key, posStdDev);
				onlineMapOfMaxs.put(key, posMax);
				onlineMapOfMins.put(key, posMin);
				onlineMapOfAmounts.put(key, posAmount);
			}
			
			HashSet<MACAddress> apSet = new HashSet<MACAddress>();
			apSet.addAll(apOfflineSet);
			apSet.addAll(apOnlineSet);
			for (MACAddress m: apSet) {
			    apChoice.addItem(m);
			}
			apChoice.addActionListener(new SelectionListener());
			apChoice.setMaximumRowCount(apChoice.getItemCount());
			JPanel apChoicePanel = new JPanel();
			apChoicePanel.setBorder(BorderFactory.createTitledBorder("Access point:"));
			GridBagLayout gdl = new GridBagLayout();
	        GridBagConstraints c = new GridBagConstraints();
	        c.fill = GridBagConstraints.BOTH;
	        c.weightx = 1.0;
	        gdl.setConstraints(apChoice, c);
	        apChoicePanel.setLayout(gdl);
			apChoicePanel.add(apChoice);
			
			/*
			// calculate the mean and the max of the rss values for different orientations
			map.clear();
			// iterate through the access points
			apIterator = mapOfMeans.keySet().iterator();
			while (apIterator.hasNext()) {
				// get the mac address
				MACAddress key = apIterator.next();
				// get the list of means for the various positions and orientations
				HashMap<GeoPosition, Double> mapOfDouble = mapOfMeans.get(key);
				// iterate through the positions and orientations
				Iterator<GeoPosition> posIterator = mapOfDouble.keySet().iterator();
				HashMap<GeoPosition, ArrayList<Double>> mapOfListOfDouble = new HashMap<GeoPosition, ArrayList<Double>>();
				while (posIterator.hasNext()) {
					// get the position and orientation
					GeoPosition pos = posIterator.next();
					// get the corresponding rss value
					double rss = mapOfDouble.get(pos);
					// create a new pos without the orientation part
					GeoPosition posWithoutOrientation = new GeoPosition(pos.getX(), pos.getY(), pos.getZ());
					ArrayList<Double> listOfDouble;
					if (mapOfListOfDouble.containsKey(posWithoutOrientation)) {
						listOfDouble = mapOfListOfDouble.get(posWithoutOrientation);
					} else {
						listOfDouble = new ArrayList<Double>();
					}
					listOfDouble.add(rss);
					mapOfListOfDouble.put(posWithoutOrientation, listOfDouble);
				}
				// mapOfListOfDouble now holds a list of the means for the different orientations on every position.
				// Iterate through this list, calculate the mean and the max for every position and
				// store these values in the means of the ap with 'max' or 'mean' as orientation identifier.
				posIterator = mapOfListOfDouble.keySet().iterator();
				while (posIterator.hasNext()) {
					GeoPosition pos = posIterator.next();
					ArrayList<Double> listOfDouble = mapOfListOfDouble.get(pos);
					mapOfDouble.put(posString + "means", Statistics.avg(listOfDouble));
					double max = -103.0;
					for (int i = 0; i < listOfDouble.size(); ++i) {
						if (listOfDouble.get(i) > max)
							max = listOfDouble.get(i);
					}
					mapOfDouble.put(posString + "max", max);
				}
			}
			*/
			
			// create the choice box for the orientations
			orientationChoice = new JComboBox();
			Collections.sort(orientationList);
			for (int i = 0; i < orientationList.size(); ++i)
				orientationChoice.addItem(orientationList.get(i));
			//orientationChoice.addItem("means");
			//orientationChoice.addItem("max");
			orientationChoice.addActionListener(new SelectionListener());
			JPanel orientationChoicePanel = new JPanel();
			orientationChoicePanel.setBorder(BorderFactory.createTitledBorder("Orientation:"));
			GridBagLayout gdlo = new GridBagLayout();
	        GridBagConstraints co = new GridBagConstraints();
	        co.fill = GridBagConstraints.BOTH;
	        co.weightx = 1.0;
	        gdlo.setConstraints(orientationChoice, co);
	        orientationChoicePanel.setLayout(gdlo);
			orientationChoicePanel.add(orientationChoice);
			
			addComponent(apChoicePanel,          4, 0, 1, 1, GridBagConstraints.HORIZONTAL);
			addComponent(orientationChoicePanel, 1, 0, 1, 1, GridBagConstraints.HORIZONTAL);
			
			
			
	        gridChoice = new JComboBox();
	        for (double d = minGridSpacing; d <= maxGridSpacing; d += gridSpacingSteps) {
	            gridChoice.addItem(d);
	        }
	        gridChoice.addActionListener(new StartPosChoiceUpdater());
	        gridChoice.addActionListener(new SelectionListener());
			JPanel gridChoicePanel = new JPanel();
			gridChoicePanel.setBorder(BorderFactory.createTitledBorder("Grid dimension:"));
			GridBagLayout gdlg = new GridBagLayout();
	        GridBagConstraints cg = new GridBagConstraints();
	        cg.fill = GridBagConstraints.BOTH;
	        cg.weightx = 1.0;
	        gdlg.setConstraints(gridChoice, cg);
	        gridChoicePanel.setLayout(gdlg);
			gridChoicePanel.add(gridChoice);
	        
	        startPosChoice = new JComboBox();
	        for (double i = 0; i < ((Double) gridChoice.getSelectedItem()); i += minGridSpacing) {
	        	for (double j = 0; j < ((Double) gridChoice.getSelectedItem()); j += minGridSpacing) {
	        		startPosChoice.addItem(new GeoPosition(i, j));
	        	}
	        }
	        startPosChoice.addActionListener(new SelectionListener());
	        JPanel startPosChoicePanel = new JPanel();
	        startPosChoicePanel.setBorder(BorderFactory.createTitledBorder("Grid start position:"));
	        gdl = new GridBagLayout();
	        c = new GridBagConstraints();
	        c.fill = GridBagConstraints.BOTH;
	        c.weightx = 1.0;
	        gdl.setConstraints(startPosChoice, c);
	        startPosChoicePanel.setLayout(gdl);
	        startPosChoicePanel.add(startPosChoice);
	        
	        gradientChoice = new JComboBox();
	        gradientChoice.addItem(RadioMapViewport.GRADIENTS_OFF);
	        gradientChoice.addItem(RadioMapViewport.GRADIENTS_DRAW_ON_HOVER);
	        gradientChoice.addItem(RadioMapViewport.GRADIENTS_ALWAYS);
	        gradientChoice.addActionListener(new SelectionListener());
			JPanel gradientChoicePanel = new JPanel();
			gradientChoicePanel.setBorder(BorderFactory.createTitledBorder("Gradient:"));
			GridBagLayout gdlgr = new GridBagLayout();
	        GridBagConstraints cgr = new GridBagConstraints();
	        cgr.fill = GridBagConstraints.BOTH;
	        cgr.weightx = 1.0;
	        gdlgr.setConstraints(gradientChoice, cgr);
	        gradientChoicePanel.setLayout(gdlgr);
			gradientChoicePanel.add(gradientChoice);
	        
	        measureChoice = new JComboBox();
	        measureChoice.addItem(RadioMapViewport.MEAN);
	        measureChoice.addItem(RadioMapViewport.STDDEV);
	        measureChoice.addItem(RadioMapViewport.MAX);
	        measureChoice.addItem(RadioMapViewport.MIN);
	        measureChoice.addItem(RadioMapViewport.AMOUNT);
	        measureChoice.addActionListener(new SelectionListener());
			JPanel measureChoicePanel = new JPanel();
			measureChoicePanel.setBorder(BorderFactory.createTitledBorder("Measure:"));
			GridBagLayout gdlm = new GridBagLayout();
	        GridBagConstraints cm = new GridBagConstraints();
	        cm.fill = GridBagConstraints.BOTH;
	        cm.weightx = 1.0;
	        gdlm.setConstraints(measureChoice, cm);
	        measureChoicePanel.setLayout(gdlm);
			measureChoicePanel.add(measureChoice);
			
			markersChoice = new JComboBox();
			markersChoice.addItem(RadioMapViewport.OFFLINE);
			markersChoice.addItem(RadioMapViewport.ONLINE);
			markersChoice.addItem(RadioMapViewport.OFFLINE_ONLINE);
			markersChoice.addItem(RadioMapViewport.NONE);
			markersChoice.addActionListener(new SelectionListener());
			JPanel markersChoicePanel = new JPanel();
			markersChoicePanel.setBorder(BorderFactory.createTitledBorder("Markers:"));
			GridBagLayout gdlma = new GridBagLayout();
	        GridBagConstraints cma = new GridBagConstraints();
	        cma.fill = GridBagConstraints.BOTH;
	        cma.weightx = 1.0;
	        gdlma.setConstraints(markersChoice, cma);
	        markersChoicePanel.setLayout(gdlma);
			markersChoicePanel.add(markersChoice);
			
			tilesChoice = new JComboBox();
			tilesChoice.addItem(RadioMapViewport.OFFLINE);
			tilesChoice.addItem(RadioMapViewport.ONLINE);
			tilesChoice.addItem(RadioMapViewport.OFFLINE_ONLINE);
			tilesChoice.addItem(RadioMapViewport.NONE);
			tilesChoice.addActionListener(new SelectionListener());
			JPanel tilesChoicePanel = new JPanel();
			tilesChoicePanel.setBorder(BorderFactory.createTitledBorder("Tiles:"));
			GridBagLayout gdlt = new GridBagLayout();
	        GridBagConstraints cmt = new GridBagConstraints();
	        cmt.fill = GridBagConstraints.BOTH;
	        cmt.weightx = 1.0;
	        gdlt.setConstraints(tilesChoice, cmt);
	        tilesChoicePanel.setLayout(gdlt);
			tilesChoicePanel.add(tilesChoice);
			
			scaleChoice = new JComboBox();
			scaleChoice.addItem(RadioMapViewport.SCALE_GLOBAL);
			scaleChoice.addItem(RadioMapViewport.SCALE_INDIVIDUAL);
			scaleChoice.addActionListener(new SelectionListener());
			JPanel scaleChoicePanel = new JPanel();
			scaleChoicePanel.setBorder(BorderFactory.createTitledBorder("Scale:"));
			GridBagLayout gdls = new GridBagLayout();
	        GridBagConstraints cs = new GridBagConstraints();
	        cs.fill = GridBagConstraints.BOTH;
	        cs.weightx = 1.0;
	        gdls.setConstraints(scaleChoice, cs);
	        scaleChoicePanel.setLayout(gdls);
			scaleChoicePanel.add(scaleChoice);
			
			apPosChoice = new JComboBox();
			apPosChoice.addItem(RadioMapViewport.AP_POS_NONE);
			apPosChoice.addItem(RadioMapViewport.AP_POS_SELECTED);
			apPosChoice.addItem(RadioMapViewport.AP_POS_ALL);
			apPosChoice.addActionListener(new SelectionListener());
	        JPanel apPosChoicePanel = new JPanel();
	        apPosChoicePanel.setBorder(BorderFactory.createTitledBorder("AP Position:"));
			GridBagLayout gdlap = new GridBagLayout();
	        GridBagConstraints cap = new GridBagConstraints();
	        cap.fill = GridBagConstraints.BOTH;
	        cap.weightx = 1.0;
	        gdlap.setConstraints(apPosChoice, cap);
	        apPosChoicePanel.setLayout(gdlap);
			apPosChoicePanel.add(apPosChoice);
			
			addComponent(gradientChoicePanel, 0, 0, 1, 1, GridBagConstraints.HORIZONTAL);
			addComponent(gridChoicePanel,     2, 0, 1, 1, GridBagConstraints.HORIZONTAL);
	        addComponent(startPosChoicePanel, 3, 0, 1, 1, GridBagConstraints.HORIZONTAL);
	        addComponent(apPosChoicePanel,    5, 0, 1, 1, GridBagConstraints.HORIZONTAL);
	        addComponent(markersChoicePanel,  6, 0, 1, 1, GridBagConstraints.HORIZONTAL);
	        addComponent(tilesChoicePanel,    7, 0, 1, 1, GridBagConstraints.HORIZONTAL);
	        addComponent(measureChoicePanel,  8, 0, 1, 1, GridBagConstraints.HORIZONTAL);
	        addComponent(scaleChoicePanel,    9, 0, 1, 1, GridBagConstraints.HORIZONTAL);
		} else if (mapMode.equals("probability")) {
			viewport.setColorBarScaleFactor(5);
			viewport.setValueScaleFactor(100);
			viewport.setValueRange(0, 1);
			viewport.setInfoMode(RadioMapViewport.INFOMODE_PROBABILITY);
			// create the choice box for the real positions
			realPosChoice = new JComboBox();
			realPosChoice.addItem(" ");
			for (int i = 0; i < realPositionList.size(); ++i) {
				realPosChoice.addItem(realPositionList.get(i));
			}
			// create the choice box for the steps
			stepChoice = new JComboBox();
			for (int i = 0; i < steps; i++) {
				stepChoice.addItem((Integer)i);
			}

			realPosChoice.addActionListener(new SelectionListener());
			JPanel realPosChoicePanel = new JPanel();
			realPosChoicePanel.setBorder(BorderFactory.createTitledBorder("Position:"));
			GridBagLayout gdlreal = new GridBagLayout();
	        GridBagConstraints creal = new GridBagConstraints();
	        creal.fill = GridBagConstraints.BOTH;
	        creal.weightx = 1.0;
	        gdlreal.setConstraints(realPosChoice, creal);
	        realPosChoicePanel.setLayout(gdlreal);
			realPosChoicePanel.add(realPosChoice);
			
			
			stepChoice.addActionListener(new SelectionListener());
			JPanel stepChoicePanel = new JPanel();
			stepChoicePanel.setBorder(BorderFactory.createTitledBorder("Step:"));
			GridBagLayout gdlstep = new GridBagLayout();
	        GridBagConstraints cstep = new GridBagConstraints();
	        cstep.fill = GridBagConstraints.BOTH;
	        cstep.weightx = 1.0;
	        gdlstep.setConstraints(stepChoice, cstep);
	        stepChoicePanel.setLayout(gdlstep);
			stepChoicePanel.add(stepChoice);
			
			addComponent(realPosChoicePanel, 0, 0, 1, 1, GridBagConstraints.HORIZONTAL);
			addComponent(stepChoicePanel,    1, 0, 1, 1, GridBagConstraints.HORIZONTAL);
		} else System.err.println("Unknown map mode");
       
        valuesCheckBox = new JCheckBox("Show values");
        valuesCheckBox.setSelected(false);
        valuesCheckBox.addChangeListener(new ValuesChangeListener());
        
		// open the window
        setTitle("RadioMap");
		//setSize(1000,600);
		setLayout(gbLayout);
		addWindowListener(new RadioMapWindowListener());
		infoText = new JLabel();
		
		JMenuBar jmb = new JMenuBar();
		JMenu jm = new JMenu("File");
		jmb.add(jm);
		JMenuItem epsExport = new JMenuItem("EPS export");
		epsExport.addActionListener(new EpsExportListener());
		jm.add(epsExport);
		jm.addSeparator();
		JMenuItem print = new JMenuItem("Print");
		print.addActionListener(new PrintListener());
		jm.add(print);
		jm.addSeparator();
		JMenuItem exit = new JMenuItem("Exit");
		exit.addActionListener(new ExitListener());
		jm.add(exit);
		setJMenuBar(jmb);

        addComponent(valuesCheckBox,      10, 0, 1, 1, GridBagConstraints.HORIZONTAL);
		addComponent(infoText,            11, 0, 1, 1, GridBagConstraints.BOTH);
		addComponent(viewport,            0, 1, 1, 12, GridBagConstraints.BOTH);

		setResizable(false);
		validate();
		pack();
		setVisible(true);
	}
	
	private void addComponent(Component c, int row, int column, int width, int height, int fill) {
        gbConstraints.gridx = column;
		gbConstraints.gridy = row;
		gbConstraints.gridwidth = width;
		gbConstraints.gridheight = height;
		gbConstraints.weightx = 1.0;
		gbConstraints.weighty = 0.0;
		gbConstraints.fill = fill;
		gbLayout.setConstraints(c, gbConstraints);
		add(c);
	}
	
	public static void main(String[] args) {
		for (int i = 0; i < args.length; i++) {
			if (args[i].equalsIgnoreCase("-offline")) {
				if (args.length <= i+1) {
					printUsage();
				}
				try {
					offlineFile = new BufferedReader(new FileReader(args[++i]));
				} catch (IOException ex) {
					System.out.println("Couldn't open the offline input file!");
					ex.printStackTrace();
					System.exit(-1);
				}
				String fileMode = getMapMode(args[i]);
				if (fileMode == null) {
					System.out.println("Offline file type unsupported.");
					printUsage();
				}
				if ((mapMode != null) && (fileMode != mapMode)) {
					System.out.println("Offline file type incompatible.");
					printUsage();
				}
				mapMode = fileMode;
			} else if (args[i].equalsIgnoreCase("-online")) {
				if (args.length <= i+1) {
					printUsage();
				}
				try {
					onlineFile = new BufferedReader(new FileReader(args[++i]));
				} catch (IOException ex) {
					System.out.println("Couldn't open the online input file!");
					ex.printStackTrace();
					System.exit(-1);
				}
				String fileMode = getMapMode(args[i]);
				if (fileMode == null) {
					System.out.println("Offline file type unsupported.");
					printUsage();
				}
				if ((mapMode != null) && (fileMode != mapMode)) {
					System.out.println("Offline file type incompatible.");
					printUsage();
				}
				mapMode = fileMode;
			} else if (args[i].equalsIgnoreCase("-maxgrid")) {
				try {
					maxGridSpacing = Double.parseDouble(args[1]);
				} catch (NumberFormatException e) {
                    printUsage();
                }
			}
		}
		if ((offlineFile == null) && (onlineFile == null)) {
			System.out.println("You must specify an offline or online file or both.");
			printUsage();
		}
		new RadioMap();
	}
    
	private static void printUsage() {
		System.out.println("Usage: RadioMap [options]");
		System.out.println("	-offline FILENAME: input file for offline data");
		System.out.println("	-online  FILENAME: input file for online data");
		System.out.println("		FILENAME: A trace file (.trace) or a probability log (.ptrace)");
		System.out.println("		(one of -offline or -online is required, both are valid)");
		System.out.println("	-maxgrid: Maximum grid spacing in meters (optional, default: 5.0)");
		System.exit(1);
	}
	
	private static String getMapMode(String file) {
		if (file.endsWith(".trace")) {
			return "tracefile";
		} else if (file.endsWith(".ptrace")) {
			return "probability";
		} else return null;
	}

	class RadioMapWindowListener extends WindowAdapter {
		public void windowClosing(WindowEvent e) {
			e.getWindow().dispose();
			System.exit(0);
		}
	}
	class StartPosChoiceUpdater implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			// reset and update startPosChoice
			synchronized(startPosChoice) {
				for (int i = startPosChoice.getItemCount(); i > 1; i--) {
					startPosChoice.removeItemAt(i - 1);
				}
				for (double i = 0; i < ((Double) gridChoice.getSelectedItem()); i += minGridSpacing) {
					for (double j = 0; j < ((Double) gridChoice.getSelectedItem()); j += minGridSpacing) {
						if (!((i == 0) && (j == 0))) { 
							startPosChoice.addItem(new GeoPosition(i, j));
						}
					}
				}
			}
		}
	}
	
	class SelectionListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			// handle tiles
			if (tilesChoice == null) {
	           	viewport.setDrawOfflineTiles(true);
	           	viewport.setDrawOnlineTiles(false);
	           	viewport.setDrawScale(false);
			} else if (tilesChoice.getSelectedItem().equals(RadioMapViewport.OFFLINE)) {
	           	viewport.setDrawOfflineTiles(true);
	           	viewport.setDrawOnlineTiles(false);
	           	viewport.setDrawScale(true);
	        } else if (tilesChoice.getSelectedItem().equals(RadioMapViewport.ONLINE)) {
	           	viewport.setDrawOfflineTiles(false);
	           	viewport.setDrawOnlineTiles(true);
	           	viewport.setDrawScale(true);
	        } else if (tilesChoice.getSelectedItem().equals(RadioMapViewport.OFFLINE_ONLINE)) {
	           	viewport.setDrawOfflineTiles(true);
	           	viewport.setDrawOnlineTiles(true);
	        } else if (tilesChoice.getSelectedItem().equals(RadioMapViewport.NONE)) {
	           	viewport.setDrawOfflineTiles(false);
	           	viewport.setDrawOnlineTiles(false);
	           	viewport.setDrawScale(false);
	        }
			// handle markers
			if (markersChoice == null) {
				viewport.setDrawOfflineMarkers(true);
				viewport.setDrawOnlineMarkers(false);
			} else if (markersChoice.getSelectedItem().equals(RadioMapViewport.OFFLINE)) {
	            viewport.setDrawOfflineMarkers(true);
	            viewport.setDrawOnlineMarkers(false);
	        } else if (markersChoice.getSelectedItem().equals(RadioMapViewport.ONLINE)) {
	           	viewport.setDrawOfflineMarkers(false);
	           	viewport.setDrawOnlineMarkers(true);
	        } else if (markersChoice.getSelectedItem().equals(RadioMapViewport.OFFLINE_ONLINE)) {
	          	viewport.setDrawOfflineMarkers(true);
	          	viewport.setDrawOnlineMarkers(true);
	        } else if (markersChoice.getSelectedItem().equals(RadioMapViewport.NONE)) {
	           	viewport.setDrawOfflineMarkers(false);
	           	viewport.setDrawOnlineMarkers(false);
	        }
	        
			double minValue = 0.0;
			double maxValue = 0.0;
			
			viewport.clearSpots();
            
			if (mapMode.equals("tracefile")) {
				minValue = 0.0;
				maxValue = -103.0;
	            double selectedGridSpacing = Double.parseDouble(gridChoice.getSelectedItem().toString());
	            GeoPosition startPos = new GeoPosition(0.0, 0.0);
	            synchronized (startPosChoice) {
	            	startPos = (GeoPosition) startPosChoice.getSelectedItem();
	            	viewport.setStartPos(startPos);
	            }
				if (!apChoice.getSelectedItem().equals(" ")) {
					
					// set position of access point
					if (apPosChoice.getSelectedItem().equals(RadioMapViewport.AP_POS_NONE)) {
						viewport.setAccessPointPosition(null);
					}
					if (apPosChoice.getSelectedItem().equals(RadioMapViewport.AP_POS_SELECTED)) {
						if (apPosition.containsKey(apChoice.getSelectedItem())) {
							Vector<GeoPosition> vector = new Vector<GeoPosition>();
							vector.add(apPosition.get(apChoice.getSelectedItem()));
							viewport.setAccessPointPosition(vector);
						} else {
							System.out.println("else");
							viewport.setAccessPointPosition(null);
						}
					}
					if (apPosChoice.getSelectedItem().equals(RadioMapViewport.AP_POS_ALL)) {
						Vector<GeoPosition> vector = new Vector<GeoPosition>();
						Set<MACAddress> keys = apPosition.keySet();
						Iterator<MACAddress> iterKeys = keys.iterator();
						while (iterKeys.hasNext()) {
							GeoPosition pos = apPosition.get(iterKeys.next());
							vector.add(pos);
						}
						viewport.setAccessPointPosition(vector);
					}
					
					// handle offline data
					Map<GeoPosition, Double> posData = new HashMap<GeoPosition, Double>();
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.MEAN))   posData = offlineMapOfMeans.get(apChoice.getSelectedItem());
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.STDDEV)) posData = offlineMapOfStdDevs.get(apChoice.getSelectedItem());
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.MAX))    posData = offlineMapOfMaxs.get(apChoice.getSelectedItem());
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.MIN))    posData = offlineMapOfMins.get(apChoice.getSelectedItem());
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.AMOUNT)) posData = offlineMapOfAmounts.get(apChoice.getSelectedItem());
					
					maxValue = Double.NEGATIVE_INFINITY;
					minValue = Double.POSITIVE_INFINITY;
					
					Iterator<GeoPosition> posIterator = posData.keySet().iterator();
					while (posIterator.hasNext()) {
						GeoPosition pos = posIterator.next();
						if ((Double.isNaN(pos.getOrientation()) && (Double.isNaN((Double) orientationChoice.getSelectedItem()))) ||
							(pos.getOrientation() == ((Double) orientationChoice.getSelectedItem()))) {
							double value = posData.get(pos);
							if (scaleChoice.getSelectedItem().equals(RadioMapViewport.SCALE_INDIVIDUAL)) {
								if (tilesChoice.getSelectedItem().equals(RadioMapViewport.OFFLINE) ||
								    tilesChoice.getSelectedItem().equals(RadioMapViewport.OFFLINE_ONLINE)) {
									if (value == 0.0) {
										System.out.println("value=0.0 " + pos);
									}
									if (value > maxValue)
										maxValue = value;
									if (value < minValue)
										minValue = value;
								}
							} else {
								if (measureChoice.getSelectedItem().equals(RadioMapViewport.MEAN)) {
									minValue = avgGlobalMinValue;
									maxValue = avgGlobalMaxValue;	 
								}
								if (measureChoice.getSelectedItem().equals(RadioMapViewport.STDDEV)) {
									minValue = stdDevGlobalMinValue;
									maxValue = stdDevGlobalMaxValue;	 
								}
								if (measureChoice.getSelectedItem().equals(RadioMapViewport.MAX)) {
									minValue = maxGlobalMinValue;
									maxValue = maxGlobalMaxValue;	 
								}
								if (measureChoice.getSelectedItem().equals(RadioMapViewport.MIN)) {
									minValue = minGlobalMinValue;
									maxValue = minGlobalMaxValue;	 
								}
								if (measureChoice.getSelectedItem().equals(RadioMapViewport.AMOUNT)) {
									minValue = amountGlobalMinValue;
									maxValue = amountGlobalMaxValue;	 
								}
							}
                            if ((pos.getX() % selectedGridSpacing == startPos.getX()) && (pos.getY() % selectedGridSpacing == startPos.getY())) {
                                viewport.addOfflineSpot(pos.getX(), pos.getY(), value);
                            }
						}
					}
					System.out.println("Viewport number of offline spots: " + viewport.getOfflineSpotSize());
										
					// handle online data
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.MEAN))   posData = onlineMapOfMeans.get(apChoice.getSelectedItem());
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.STDDEV)) posData = onlineMapOfStdDevs.get(apChoice.getSelectedItem());
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.MAX))    posData = onlineMapOfMaxs.get(apChoice.getSelectedItem());
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.MIN))    posData = onlineMapOfMins.get(apChoice.getSelectedItem());
					if (measureChoice.getSelectedItem().equals(RadioMapViewport.AMOUNT)) posData = onlineMapOfAmounts.get(apChoice.getSelectedItem());
					
					if (posData != null) {
						posIterator = posData.keySet().iterator();
						while (posIterator.hasNext()) {
							GeoPosition pos = posIterator.next();
							if ((Double.isNaN(pos.getOrientation()) && (Double.isNaN((Double) orientationChoice.getSelectedItem()))) ||
							    (pos.getOrientation() == ((Double) orientationChoice.getSelectedItem()))) {
								double value = posData.get(pos);
								if (scaleChoice.getSelectedItem().equals(RadioMapViewport.SCALE_INDIVIDUAL)) {
									if (tilesChoice.getSelectedItem().equals(RadioMapViewport.ONLINE) ||
									    tilesChoice.getSelectedItem().equals(RadioMapViewport.OFFLINE_ONLINE)) {
										if (value == 0.0) {
											System.out.println("value=0.0 " + pos);
										}
										if (value > maxValue)
											maxValue = value;
										if (value < minValue)
											minValue = value;
									}
								} else {
									if (measureChoice.getSelectedItem().equals(RadioMapViewport.MEAN)) {
										minValue = avgGlobalMinValue;
										maxValue = avgGlobalMaxValue;	 
									}
									if (measureChoice.getSelectedItem().equals(RadioMapViewport.STDDEV)) {
										minValue = stdDevGlobalMinValue;
										maxValue = stdDevGlobalMaxValue;	 
									}
									if (measureChoice.getSelectedItem().equals(RadioMapViewport.MAX)) {
										minValue = maxGlobalMinValue;
										maxValue = maxGlobalMaxValue;	 
									}
									if (measureChoice.getSelectedItem().equals(RadioMapViewport.MIN)) {
										minValue = minGlobalMinValue;
										maxValue = minGlobalMaxValue;	 
									}
									if (measureChoice.getSelectedItem().equals(RadioMapViewport.AMOUNT)) {
										minValue = amountGlobalMinValue;
										maxValue = amountGlobalMaxValue;	 
									}
								}
								viewport.addOnlineSpot(pos.getX(), pos.getY(), value);
							}
						}
					}
					System.out.println("Viewport number of online spots: " + viewport.getOnlineSpotSize());
					
				} else {
					minValue = 0;
					maxValue = 0;
					viewport.setAccessPointPosition(null);
				}
	            viewport.setDrawGradients(gradientChoice.getSelectedItem().toString());        
	            viewport.setTileSize(selectedGridSpacing);
			} else if (mapMode.equals("probability")) {
				if (!realPosChoice.getSelectedItem().equals(" ")) {
					minValue = 1.0;
					maxValue = 0.0;
					GeoPosition realPos = (GeoPosition) realPosChoice.getSelectedItem();
					Integer step = (Integer)stepChoice.getSelectedItem();
					Map<GeoPosition, Double> mapOfListOfDouble = offlineRiceProbabilities.get(new GeoPositionStep(realPos, step));
					viewport.showCross(realPos.getX(), realPos.getY());
					Iterator<GeoPosition> posIterator = mapOfListOfDouble.keySet().iterator();
					while (posIterator.hasNext()) {
						GeoPosition pos = posIterator.next();
						double value = mapOfListOfDouble.get(pos);
						/*if (scaleChoice.getSelectedItem().equals("individual scale")) {
							if (value > maxValue)
								maxValue = value;
							if (value < minValue)
								minValue = value;
						} else {*/
							minValue = avgGlobalMinValue;
							maxValue = avgGlobalMaxValue;
						//}
                        viewport.addOfflineSpot(pos.getX(), pos.getY(), value);
					}
				} else {
					minValue = 0.0;
					maxValue = 1.0;
					viewport.hideCross();
				}
				viewport.setTileSize(1.0);
			}
			viewport.setValueRange(minValue, maxValue);
			viewport.repaint();
		}
	}
	
	class ValuesChangeListener implements ChangeListener {
		public void stateChanged(ChangeEvent e) {
			if (e.getSource().equals(valuesCheckBox)) {
				viewport.setDrawValues(valuesCheckBox.isSelected());
				viewport.repaint();
			}
		}
	}
	
	class EpsExportListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			try {
				JFileChooser jfc = new JFileChooser();
				//FileNameExtensionFilter filter = new FileNameExtensionFilter("EPS Images", "eps");
				//jfc.setFileFilter(filter);
				jfc.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int returnVal = jfc.showSaveDialog(RadioMap.this);
			    if (returnVal == JFileChooser.APPROVE_OPTION) {
			    	EpsGraphics2D g2d = new EpsGraphics2D();
			    	viewport.print(g2d);
			    	FileWriter fw = new FileWriter(jfc.getSelectedFile());
			    	fw.write(g2d.toString());
			    	fw.flush();
			    	fw.close();
			    }
			} catch (IOException ioe) {
				ioe.printStackTrace();
                JOptionPane.showMessageDialog(RadioMap.this, ioe, "Error", JOptionPane.ERROR_MESSAGE);
			}
		}
	}
	
	class PrintListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			viewport.print();
		}
	}
	
	class ExitListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			System.exit(0);
		}
	}
}
