/*
 * This file is part of Locana.
 * 
 * Copyright (c) 2007 Thomas King <king@informatik.uni-mannheim.de>,
 * University of Mannheim, Germany
 * 
 * All rights reserved.
 *
 * Loclib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Loclib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Loclib; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package org.pi4.locana.radiomap;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.Rectangle2D;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import javax.print.PrintService;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.OrientationRequested;
import javax.swing.JPanel;

import org.apache.batik.swing.JSVGCanvas;
import org.pi4.locutil2.GeoPosition;

@SuppressWarnings("serial")
public class RadioMapViewport extends JPanel {
    public static final String GRADIENTS_DRAW_ON_HOVER = "on hover";
    public static final String GRADIENTS_OFF = "off";
    public static final String GRADIENTS_ALWAYS = "always";
    public static final String SCALE_INDIVIDUAL = "individual";
    public static final String SCALE_GLOBAL = "global";
    public static final String MEAN = "mean";
    public static final String STDDEV = "stdDev";
    public static final String MAX = "max";
    public static final String MIN = "min";
    public static final String AMOUNT = "amount";
    public static final String INFOMODE_TRACE = "trace";
    public static final String INFOMODE_PROBABILITY = "probability";
    public static final String OFFLINE = "offline";
    public static final String ONLINE = "online";
    public static final String OFFLINE_ONLINE = "both";
    public static final String AP_POS_NONE = "none";
    public static final String AP_POS_SELECTED = "selected";
    public static final String AP_POS_ALL = "all";
    public static final String NONE = "none";

    private static final double FACTOR_M_TO_PX = 29.19;
    private static final double ORIGIN_X = 27;
    private static final double ORIGIN_Y = 428;
    private static final int COLORBAR_Y = 580;
    private static final int COLORBAR_MINMAXSTRING_Y = 595;
    private static final int COLORBAR_TOPSTRING_Y = 570;
    private static final int INFO_X = 25;
    private static final int INFO1_Y = 690;
    private static final int INFO2_Y = 720;
    private static final int INFO3_Y = 750;
    private static final int GRADIENTS_MARGIN_X_X = 2;
    private static final int GRADIENTS_MARGIN_X_Y = 2;
    private static final int GRADIENTS_MARGIN_Y_X = -3;
    private static final int GRADIENTS_MARGIN_Y_Y = -4;
    private static final int INFO_FONTSIZE = 14;
    private static final String BACKGROUND_IMAGE = "res/GroundPlan.svg";
    private static final Dimension CANVAS_PREFERRED_SIZE = new Dimension(1024, 533);
    
	//private List<Color> colorList;
	private double minValue, maxValue;
	private double tileSize; // tile size in pixels
    private double originX;
    private double originY;
    private int colorBarY;
    private int colorBarMinMaxStringY;
    private int colorBarTopStringY;
    private int infoX;
    private int info1Y;
    private int info2Y;
    private int info3Y;
    private boolean drawOfflineMarkers;
    private boolean drawOnlineMarkers;
    private boolean drawOfflineTiles;
    private boolean drawOnlineTiles;
    private boolean drawValues;
    private boolean drawScale;
    private boolean gradientsAlways;
    private boolean gradientsDrawOnHover;
	private String info1 = "", info2 = "", info3 = "";
	private Map<GeoPosition, Double> offlineSpotList, onlineSpotList;
	private double pointedX, pointedY;
	private int colorBarScaleFactor, valueScaleFactor;
	private boolean showCross = false;
	private double crossX = -100.0, crossY = -100.0;
	private String infoMode = INFOMODE_TRACE;
	private GeoPosition startPos = new GeoPosition(0.0, 0.0);
	private Vector<GeoPosition> apPosition;
    
    private JSVGCanvas svgCanvas;
	
	public RadioMapViewport() {
		this.setBorder(null);
		setBackground(Color.WHITE);
		setForeground(Color.BLACK);
		setOpaque(true);
        setPreferredSize(new Dimension(1024, 768));
        setMinimumSize(new Dimension(1024, 768));
        setMaximumSize(new Dimension(1024, 768));
        setAlignmentX(0.0f);
        setAlignmentY(0.0f);
        try {
            URL imageFile = new File(BACKGROUND_IMAGE).toURI().toURL();
            if (!(new File(BACKGROUND_IMAGE).exists())) {
            	imageFile = ClassLoader.getSystemResource(BACKGROUND_IMAGE);
            	//if (!imageFile.exists()) throw new FileNotFoundException("Ground plan image file not found");
            }
            svgCanvas = new JSVGCanvas();
            svgCanvas.setPreferredSize(CANVAS_PREFERRED_SIZE);
            svgCanvas.setMinimumSize(CANVAS_PREFERRED_SIZE);
			svgCanvas.setMaximumSize(CANVAS_PREFERRED_SIZE);
            svgCanvas.setURI(imageFile.toString());
            svgCanvas.setForeground(Color.BLACK);
            svgCanvas.setBackground(Color.WHITE);
            svgCanvas.setOpaque(true);
            svgCanvas.addMouseMotionListener(new RadioMapViewportMouseMotionListener());
            svgCanvas.addMouseListener(new RadioMapViewportMouseListener());
            add(svgCanvas);
        } catch (MalformedURLException e) {
            e.printStackTrace();
        }
        
        originX = ORIGIN_X;
        originY = ORIGIN_Y;
        colorBarY = COLORBAR_Y;
        colorBarMinMaxStringY = COLORBAR_MINMAXSTRING_Y;
        colorBarTopStringY = COLORBAR_TOPSTRING_Y;
        infoX = INFO_X;
        info1Y = INFO1_Y;
        info2Y = INFO2_Y;
        info3Y = INFO3_Y;
        drawOfflineMarkers = true;
        drawOnlineMarkers = false;
        drawOfflineTiles = true;
        drawOnlineTiles = false;
        drawValues = false;
        drawScale = false;
        gradientsAlways = false;
        gradientsDrawOnHover = false;
		offlineSpotList = new HashMap<GeoPosition, Double>();
		onlineSpotList  = new HashMap<GeoPosition, Double>();
		setVisible(true);
	}
	
	public void print(Graphics2D g2d) {
		synchronized(this) {
			printOrPaint(g2d);
			// paint svg
			svgCanvas.setPreferredSize(CANVAS_PREFERRED_SIZE);
			//svgCanvas.setMinimumSize(CANVAS_PREFERRED_SIZE);
			//svgCanvas.setMaximumSize(CANVAS_PREFERRED_SIZE);
			svgCanvas.getGraphicsNode().paint(g2d);
			
		}
	}
	
	public void paint(Graphics g) {
		synchronized(this) {
			// whitening the background
			g.setColor(Color.WHITE);
			g.fillRect(0, 0, 1024, 768);

			// paint svg
			svgCanvas.setPreferredSize(CANVAS_PREFERRED_SIZE);
			svgCanvas.setMinimumSize(CANVAS_PREFERRED_SIZE);
			svgCanvas.setMaximumSize(CANVAS_PREFERRED_SIZE);
			svgCanvas.paint(g);
			printOrPaint(g);
		}
	}
	
    public void printOrPaint(Graphics g) {
		int colorBarRight = 960;
		int colorBarWidth = (getScaledCeiling(maxValue) - getScaledFloor(minValue)) * colorBarScaleFactor;
		int colorBarLeft = colorBarRight - colorBarWidth;
                
		/*for (int i = 0; i < colorList.size(); i++) {
			if (i == 0) {
				String s = Double.toString(Math.round(minValue * 10.0) / 10.0);
				Rectangle2D rect = g.getFontMetrics().getStringBounds(s, g);
				g.setColor(Color.BLACK);
				g.drawString(s, colorBarLeft + (i * colorBarScaleFactor) - (int) rect.getMaxX() - 10, colorBarMinMaxStringY);
			} else if (i == (colorList.size() / 2)) {
				String s = Double.toString(Math.round((minValue + maxValue) * 50.0) / 100.0);
				Rectangle2D rect = g.getFontMetrics().getStringBounds(s, g);
				g.setColor(Color.BLACK);
				g.drawString(s, colorBarLeft + (i * colorBarScaleFactor) - (int) rect.getCenterX(), colorBarTopStringY);
			} else if (i == colorList.size() - 1) {
				String s = Double.toString(Math.round(maxValue * 10.0) / 10.0);
				g.setColor(Color.BLACK);
				g.drawString(s, colorBarLeft + (i * colorBarScaleFactor) + 10, colorBarMinMaxStringY);
			}
			g.setColor(colorList.get(i));
			g.fillRect(colorBarLeft + (i * colorBarScaleFactor), colorBarY, colorBarScaleFactor, 20);
		}*/
		
		// draw the scale
		if (drawScale) {
			if (maxValue > minValue) {
				Font font = g.getFont();
				Font newFont = font.deriveFont(font.getSize() + 6.0f);
				g.setFont(newFont);
				String s = Double.toString(Math.round(minValue * 10.0) / 10.0);
				Rectangle2D rect = g.getFontMetrics().getStringBounds(s, g);
				g.setColor(Color.BLACK);
				g.drawString(s, colorBarLeft - (int) rect.getMaxX() - 10, colorBarMinMaxStringY);
				s = Double.toString(Math.round((minValue + maxValue) * 50.0) / 100.0);
				rect = g.getFontMetrics().getStringBounds(s, g);
				g.setColor(Color.BLACK);
				g.drawString(s, colorBarLeft + (colorBarWidth/2) - (int) rect.getCenterX(), colorBarTopStringY);
				s = Double.toString(Math.round(maxValue * 10.0) / 10.0);
				g.setColor(Color.BLACK);
				g.drawString(s, colorBarLeft + colorBarWidth + 10, colorBarMinMaxStringY);

				for (int i = 0; i <= colorBarWidth; i++) {
					double d = (double) i / (double)colorBarWidth;
					d *= (maxValue - minValue);
					d *= valueScaleFactor;
					g.setColor(getColor(d));
					g.fillRect(colorBarLeft + i, colorBarY, 1, 20);
				}
			}
		}
		
		// draw the tiles
		Iterator<GeoPosition> posIterator = offlineSpotList.keySet().iterator();
		while (posIterator.hasNext()) {
			GeoPosition pos = posIterator.next();
			Double spot = offlineSpotList.get(pos);
			double centerX = originX + (pos.getX() * FACTOR_M_TO_PX);
			double centerY = originY - (pos.getY() * FACTOR_M_TO_PX);
			int left   = new Long(Math.round(centerX - (tileSize / 2.0))).intValue();
			int right  = new Long(Math.round(centerX + (tileSize / 2.0))).intValue();
			int top    = new Long(Math.round(centerY - (tileSize / 2.0))).intValue();
			int bottom = new Long(Math.round(centerY + (tileSize / 2.0))).intValue();
			// draw offline tiles
			if (drawOfflineTiles) {
				g.setColor(getColor((spot - minValue) * valueScaleFactor));
				g.fillRect(left, top, right - left, bottom - top);
			}
			else {
				g.setColor(Color.LIGHT_GRAY);
				g.fillRect(left, top, right - left, bottom - top);
			}
            // draw offline markers
            if (drawOfflineMarkers) {
                g.setColor(Color.BLUE);
                g.fillRect((int) Math.round(centerX - 2.0), (int) Math.round(centerY - 2.0), 5, 5);
                //g.drawLine((int) Math.round(centerX - 1.0), (int) Math.round(centerY), (int) Math.round(centerX),
                //		   (int) Math.round(centerY));
                //g.drawLine((int) Math.round(centerX), (int) Math.round(centerY), (int) Math.round(centerX),
                //		   (int) Math.round(centerY + 1.0));
            }
            // draw values
            if (drawValues && (drawOfflineMarkers || drawOfflineTiles)) {
            	g.setColor(Color.BLACK);
            	g.setFont(new Font(null, Font.TRUETYPE_FONT, 8));
            	int temp = (int) Math.round(spot);
            	drawNumber(temp, (int) Math.round(centerX - 5.0), (int) Math.round(centerY), true, g);
            }
		}
		
		posIterator = onlineSpotList.keySet().iterator();
		while (posIterator.hasNext()) {
			GeoPosition pos = posIterator.next();
			Double spot = onlineSpotList.get(pos);
			double centerX = originX + (pos.getX() * FACTOR_M_TO_PX);
			double centerY = originY - (pos.getY() * FACTOR_M_TO_PX);
			int left = new Long(Math.round(centerX - (tileSize / 2.0))).intValue();
			int right = new Long(Math.round(centerX + (tileSize / 2.0))).intValue();
			int top = new Long(Math.round(centerY - (tileSize / 2.0))).intValue();
			int bottom = new Long(Math.round(centerY + (tileSize / 2.0))).intValue();
			// draw online tiles
			if (drawOnlineTiles) {
				g.setColor(getColor((spot - minValue) * valueScaleFactor));
				g.fillRect(left, top, right - left, bottom - top);
			}
            // draw online markers
            if (drawOnlineMarkers) {
                g.setColor(new Color(255, 0, 175));
                g.fillRect((int) Math.round(centerX - 2.0), (int) Math.round(centerY - 2.0), 5, 5);
                //g.drawLine((int) Math.round(centerX - 1.0), (int) Math.round(centerY + 1.0), (int) Math.round(centerX),
                //		   (int) Math.round(centerY + 1.0));
                //g.drawLine((int) Math.round(centerX), (int) Math.round(centerY), (int) Math.round(centerX),
                //		   (int) Math.round(centerY + 1.0));
            }
            // draw values
            if (drawValues && (drawOnlineMarkers || drawOnlineTiles)) {
            	g.setColor(Color.BLACK);
            	g.setFont(new Font(null, Font.TRUETYPE_FONT, 8));
            	int temp = (int) Math.round(spot);
            	drawNumber(temp, (int) Math.round(centerX - 5.0), (int) Math.round(centerY), true, g);
            }
		}
            
        // draw gradients on the tile that is pointed at with the mouse (if "draw on hover" is selected)
        if (gradientsDrawOnHover) {
        	if (offlineSpotList.containsKey(new GeoPosition(pointedX, pointedY, 0.0))) {
        		Double hoverSpot = offlineSpotList.get(new GeoPosition(pointedX, pointedY, 0.0));
                double hoverSpotCenterX = originX + (pointedX * FACTOR_M_TO_PX);
                double hoverSpotCenterY = originY - (pointedY * FACTOR_M_TO_PX);
                double prec = tileSize / FACTOR_M_TO_PX;
                double nextX = pointedX + prec;
                double nextY = pointedY + prec;
                // gradient in x-direction
                if (offlineSpotList.containsKey(new GeoPosition(nextX, pointedY, 0.0))) {
                	double nextValue = offlineSpotList.get(new GeoPosition(nextX, pointedY, 0.0));
                	int diff = (int) Math.round(nextValue - hoverSpot);
                	g.setColor(Color.BLACK);
                	drawNumber(diff, (int) Math.round(hoverSpotCenterX + GRADIENTS_MARGIN_X_X), 
                			   (int) Math.round(hoverSpotCenterY + GRADIENTS_MARGIN_X_Y), true, g);
                }
                // gradient in y-direction
                if (offlineSpotList.containsKey(new GeoPosition(pointedX, nextY, 0.0))) {
                	double nextValue = offlineSpotList.get(new GeoPosition(pointedX, nextY, 0.0));
                	int diff = (int) Math.round(nextValue - hoverSpot);
                	g.setColor(Color.BLACK);
                	drawNumber(diff, (int) Math.round(hoverSpotCenterX + GRADIENTS_MARGIN_Y_X), 
                			   (int) Math.round(hoverSpotCenterY + GRADIENTS_MARGIN_Y_Y), true, g);
                    	
                }
            }
		}
        
		// draw gradients on each tile if "always" is selected in gradientChoice
        // (must be done with own iterator, because otherwise the gradients are partly 
        // "overpainted" by adjacent tiles)
        if (gradientsAlways) {
            posIterator = offlineSpotList.keySet().iterator();
            while (posIterator.hasNext()) {
            	GeoPosition pos = posIterator.next();
                Double spot = offlineSpotList.get(pos);
                double centerX = originX + (pos.getX() * FACTOR_M_TO_PX);
                double centerY = originY - (pos.getY() * FACTOR_M_TO_PX);
                double prec = tileSize / FACTOR_M_TO_PX; // grid dimension in meters
                double nextX = pos.getX() + prec;
                double nextY = pos.getY() + prec;
                // gradient in x-direction
                if (offlineSpotList.containsKey(new GeoPosition(nextX, pos.getY(), 0.0))) {
                    double nextValue = offlineSpotList.get(new GeoPosition(nextX, pos.getY(), 0.0));
                    int diff = (int) Math.round(nextValue - spot);
                    g.setColor(Color.BLACK);
                    drawNumber(diff, (int) Math.round(centerX + GRADIENTS_MARGIN_X_X),
                    		   (int) Math.round(centerY + GRADIENTS_MARGIN_X_Y), true, g);
                }
                // gradient in y-direction
                if (offlineSpotList.containsKey(new GeoPosition(pos.getX(), nextY, 0.0))) {
                    double nextValue = offlineSpotList.get(new GeoPosition(pos.getX(), nextY, 0.0));
                    int diff = (int) Math.round(nextValue - spot);
                    g.setColor(Color.BLACK);
                    drawNumber(diff, (int) (Math.round(centerX + GRADIENTS_MARGIN_Y_X)), 
                               (int) Math.round(centerY + GRADIENTS_MARGIN_Y_Y), true, g);
                }
            }
        }
		
		// draw the cross, if necessary
		if (showCross) {
			double centerX = originX + (crossX * FACTOR_M_TO_PX);
			double centerY = originY - (crossY * FACTOR_M_TO_PX);
			int left = (int) Math.round(centerX - 5.0);
			int right = (int) Math.round(centerX + 5.0);
			int top = (int) Math.round(centerY - 5.0);
			int bottom = (int) Math.round(centerY + 5.0);
			g.setColor(Color.BLUE);
			g.drawLine(left, top, right, bottom);
			g.drawLine(left, bottom, right, top);
		}
		
		// draw position of access point
		if (apPosition != null) {
			g.setColor(Color.ORANGE);
			for (GeoPosition pos: apPosition) {
				if (pos.getZ() == 0.0) g.fillOval((int) pos.getX(), (int) pos.getY(), 10, 10);
				if (pos.getZ() != 0.0) g.drawOval((int) pos.getX(), (int) pos.getY(), 10, 10);
			}
		}
		
		// draw the info lines
		g.setColor(Color.BLACK);
        g.setFont(new Font(null, Font.TRUETYPE_FONT, INFO_FONTSIZE));
		g.drawString(info1, infoX, info1Y);
		g.drawString(info2, infoX, info2Y);
		g.drawString(info3, infoX, info3Y);
	}
    
    public void drawNumber(int number, int x, int y, boolean signed, Graphics g) {
        if (signed) {
            g.drawLine(x, y, x + 2, y);
            if (number >= 0) {
                g.drawLine(x + 1, y - 1, x + 1, y + 1);
            }
            drawNumber(Math.abs(number), x + 4, y, false, g);
            return;
        } else {
            int digit = Math.abs(number);
            while (digit > 9) {
                digit = (int) (digit / 10);
            }
            switch (digit) {
                case 0: {
                    g.drawLine(x, y - 2, x, y + 2);
                    g.drawLine(x, y + 2, x + 2, y + 2);
                    g.drawLine(x, y - 2, x + 2, y - 2);
                    g.drawLine(x + 2, y - 2, x + 2, y + 2);
                }
                break;
                case 1: {
                    g.drawLine(x, y + 2, x + 2, y + 2);
                    g.drawLine(x, y - 2, x + 1, y - 2);
                    g.drawLine(x + 1, y - 2, x + 1, y + 2);
                }
                break;
                case 2: {
                    g.drawLine(x, y, x + 2, y);
                    g.drawLine(x, y - 2, x + 2, y - 2);
                    g.drawLine(x, y + 2, x + 2, y + 2);
                    g.drawLine(x, y + 1, x, y + 1);
                    g.drawLine(x + 2, y - 1, x + 2, y - 1);
                }
                break;
                case 3: {
                    g.drawLine(x, y - 2, x + 2, y - 2);
                    g.drawLine(x, y + 2, x + 2, y + 2);
                    g.drawLine(x, y, x + 2, y);
                    g.drawLine(x + 2, y - 2, x + 2, y + 2);
                }
                break;
                case 4: {
                    g.drawLine(x, y - 2, x, y);
                    g.drawLine(x + 2, y - 2, x + 2, y + 2);
                    g.drawLine(x + 1, y, x + 1, y);
                }
                break;
                case 5: {
                    g.drawLine(x, y - 2, x + 2, y - 2);
                    g.drawLine(x, y, x + 2, y);
                    g.drawLine(x, y + 2, x + 2, y + 2);
                    g.drawLine(x, y - 1, x, y - 1);
                    g.drawLine(x + 2, y + 1, x + 2, y + 1);
                }
                break;
                case 6: {
                    g.drawLine(x, y - 2, x + 2, y - 2);
                    g.drawLine(x, y - 2, x, y + 2);
                    g.drawLine(x, y, x + 2, y);
                    g.drawLine(x, y + 2, x + 2, y + 2);
                    g.drawLine(x + 2, y + 1, x + 2, y + 1);
                }
                break;
                case 7: {
                    g.drawLine(x, y - 2, x + 2, y - 2);
                    g.drawLine(x + 2, y - 2, x + 2, y + 2);
                }
                break;
                case 8: {
                    g.drawLine(x, y - 2, x, y + 2);
                    g.drawLine(x + 2, y - 2, x + 2, y + 2);
                    g.drawLine(x + 1, y - 2, x + 1, y - 2);
                    g.drawLine(x + 1, y, x + 1, y);
                    g.drawLine(x + 1, y + 2, x + 1, y + 2);
                }
                break;
                case 9: {
                    g.drawLine(x, y - 2, x + 2, y - 2);
                    g.drawLine(x + 2, y - 2, x + 2, y + 2);
                    g.drawLine(x, y - 2, x, y);
                    g.drawLine(x + 1, y, x + 1, y);
                    g.drawLine(x, y + 2, x + 2, y + 2);
                }
                break;
                default: {
                    throw new IllegalArgumentException();
                }
            }
            if (Math.abs(number) > 999) {
                throw new IllegalArgumentException("Number out of range; |n| must be <1000");
            }
            if (Math.abs(number) > 99) {
                drawNumber(Math.abs(number % 100), x + 4, y, false, g);
            } else if (Math.abs(number) > 9) {
                drawNumber(Math.abs(number % 10), x + 4, y, false, g);
            }
        }
    }
	
	public void showCross(double x, double y) {
		showCross = true;
		crossX = x;
		crossY = y;
	}
	
	public void hideCross() {
		showCross = false;
	}
	
	public void setColorBarScaleFactor(int sf) {
		colorBarScaleFactor = sf;
	}
	
	public void setValueScaleFactor(int sf) {
		valueScaleFactor = sf;
	}
	
	public void setValueRange(double low, double high) {
		minValue = low;
		maxValue = high;
		/*double difference = getScaledCeiling(high) - getScaledFloor(low);
		colorList = new ArrayList<Color>();
		for (int i = 0; i <= difference; i++) {
			int r = 512 - new Long(Math.round(i * 512.0 / difference)).intValue();
			if (r < 0) r = 0;
			if (r > 255) r = 255;
			int g = new Long(Math.round(i * 512.0 / difference)).intValue();
			if (g > 255) g = 255;
			if (g < 0) g = 0;
			colorList.add(new Color(r, g, 0, 255));
		}*/
	}
	
	private int getScaledCeiling(double value) {
		return new Double(Math.ceil(value * new Double(valueScaleFactor))).intValue();
	}
	
	private int getScaledFloor(double value) {
		return new Double(Math.floor(value * new Double(valueScaleFactor))).intValue();
	}
	
	private Color getColor(double value) {
		double difference = maxValue * valueScaleFactor - minValue * valueScaleFactor;
		int r = 512 - new Long(Math.round(value * valueScaleFactor * 512.0 / difference)).intValue();
		if (r < 0)
			r = 0;
		if (r > 255)
			r = 255;
		int g = new Long(Math.round(value * valueScaleFactor * 512.0 / difference)).intValue();
		if (g > 255)
			g = 255;
		if (g < 0)
			g = 0;

		return new Color(r, g, 0, 128);
		
		/*double range = maxValue * valueScaleFactor - minValue * valueScaleFactor;
		//value = Math.log(value/range)/6 + 1.0f;
		//value = 64.0 * Math.pow(value/range - 0.5, 7) + 0.5;
		value = Math.exp(4.0 * (value/range - 1.0));
		value *= 2.0;
		float r = (float)(2.0 - value);
		float g = (float)(value);
		if (r > 1.0f)
			r = 1.0f;
		if (g > 1.0f)
			g = 1.0f;
		if (r < 0.0f)
			r = 0.0f;
		if (g < 0.0f)
			g = 0.0f;
		return new Color(r, g, 0.0f, 0.5f);*/
	}
	
	public void setTileSize(double sizeInMeters) {
        tileSize = sizeInMeters * FACTOR_M_TO_PX;
    }
	
	public void setDrawOfflineMarkers(boolean value) {
        drawOfflineMarkers = value;
    }
	
	public void setDrawOnlineMarkers(boolean value) {
        drawOnlineMarkers = value;
    }
	
	public void setDrawOfflineTiles(boolean value) {
        drawOfflineTiles = value;
    }
	
	public void setDrawOnlineTiles(boolean value) {
        drawOnlineTiles = value;
    }
	
	public void setDrawValues(boolean value) {
		drawValues = value;
	}
    
    public void setDrawGradients(String value) {
        if (value.equals(GRADIENTS_ALWAYS)) {
            gradientsAlways = true;
            gradientsDrawOnHover = false;
        } else if (value.equals(GRADIENTS_DRAW_ON_HOVER)) {
            gradientsAlways = false;
            gradientsDrawOnHover = true;
        } else if (value.equals(GRADIENTS_OFF)) {
            gradientsAlways = false;
            gradientsDrawOnHover = false;
        } else {
            throw new IllegalArgumentException();
        }
    }
    
	public void setDrawScale(boolean value) {
		drawScale = value;
	}
    
    public void setAccessPointPosition(Vector<GeoPosition> pos) {
    	apPosition = pos;
    }
	
	public double getPointedX() {
		return pointedX;
	}
	
	public double getPointedY() {
		return pointedY;
	}
	
	public void setInfoMode(String mode) {
		if (mode.equals(INFOMODE_TRACE) || mode.equals(INFOMODE_PROBABILITY))
            infoMode = mode;
        else
            throw new IllegalArgumentException("Unsupported info mode");
	}
	
	public void addOfflineSpot(double x, double y, double value) {
		offlineSpotList.put(new GeoPosition(x, y, 0.0), value);
	}
	
	public int getOfflineSpotSize() {
		return offlineSpotList.size();
	}
	
	public void addOnlineSpot(double x, double y, double value) {
		onlineSpotList.put(new GeoPosition(x, y, 0.0), value);
	}
	
	public int getOnlineSpotSize() {
		return onlineSpotList.size();
	}

	
	public void clearSpots() {
		offlineSpotList.clear();
		onlineSpotList.clear();
		info1 = info2 = info3 = "";
	}
	
	public void setStartPos(GeoPosition startPos) {
		this.startPos = startPos;
	}
	
	public void print() {
		new Print2DPrinterJob();
	}
	
	public void printComponents(Graphics g) {
		System.out.println("printComponents");
		paintComponents(g);
	}
	
	class RadioMapViewportMouseMotionListener extends MouseMotionAdapter {
		public void mouseMoved(MouseEvent e) {
			//System.out.println("x, y: " + e.getX() + ", " + e.getY());
			double x = (e.getX() - originX) / FACTOR_M_TO_PX; // x in meters relative to null point of the measurement point grid 
			double y = (originY - e.getY())  / FACTOR_M_TO_PX; // y in meters relative to null point of the measurement point grid
            double prec = tileSize / FACTOR_M_TO_PX; // grid dimension in meters
            double modolo = (x % prec) - startPos.getX();
            if (modolo > (0.5 * prec)) x = x + (prec - modolo);
            if (modolo <= (0.5 * prec)) x = x - modolo;
            modolo = (y % prec) - startPos.getY();
            if (modolo > (0.5 * prec)) y = y + (prec - modolo);
            if (modolo <= (0.5 * prec)) y = y - modolo;
			if ((x != pointedX) || (y != pointedY)) {
				pointedX = x;
				pointedY = y;
                if (offlineSpotList.containsKey(new GeoPosition(x, y, 0.0))) {
					info1 = "x   = " + x;
					info2 = "y   = " + y;
					if (infoMode.equals(INFOMODE_TRACE))
						info3 = "value = " + (Math.round(offlineSpotList.get(new GeoPosition(x, y, 0.0)) * 100.0) / 100.0);
					else if (infoMode.equals(INFOMODE_PROBABILITY))
						info3 = "p = " + offlineSpotList.get(new GeoPosition(x, y, 0.0));
				} else {
					info1 = info2 = info3 = "";
				}
				repaint();
			}
		}
	}
	
	class RadioMapViewportMouseListener extends MouseAdapter {
		public void mouseExited(MouseEvent e) {
			info1 = info2 = info3 = "";
			repaint();
		}
	}
	
	public class Print2DPrinterJob implements Printable {

		public Print2DPrinterJob() {
			PrintRequestAttributeSet aset = new HashPrintRequestAttributeSet();
			aset.add(OrientationRequested.LANDSCAPE);
			
			PrinterJob pj = PrinterJob.getPrinterJob();       
			pj.setPrintable(this);
			PrintService[] services = PrinterJob.lookupPrintServices();

			if (services.length > 0) {
				try {
					pj.setPrintService(services[0]);
					if (pj.printDialog(aset)) {
						pj.print(aset);
					}
				} catch (PrinterException pe) {
					System.err.println(pe);
				}
			}
		}

		public int print(Graphics g, PageFormat pf, int pageIndex) {
			if (pageIndex == 0) {
				Graphics2D g2d = (Graphics2D) g;
				g2d.translate(pf.getImageableX(), pf.getImageableY());
                g2d.scale(0.6, 0.6);
				RadioMapViewport.this.paint(g2d);
				return Printable.PAGE_EXISTS;					
			} else {
				return Printable.NO_SUCH_PAGE;
			}
		}
	}
}
