/*
 * This file is part of Locana.
 * 
 * Copyright (c) 2007 Thomas King <king@informatik.uni-mannheim.de>,
 * University of Mannheim, Germany
 * 
 * All rights reserved.
 *
 * Loclib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Loclib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Loclib; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package org.pi4.locana.radiomap;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.pi4.locutil1.MACAddress;
import org.pi4.locutil2.GeoPosition;

public class TraceAnalyzer {
	
	public static void analyzeTracefile(BufferedReader file, Map<MACAddress,Map<GeoPosition,List<Double>>> macSignalStrengths, Map<MACAddress,GeoPosition> apPositions, List<Double> orientations) throws IOException {
		double degmap = 45;
		double x = Double.NaN, y = Double.NaN, degree = Double.NaN;
		String line;
		while ((line = file.readLine()) != null) {
			// handle position of access points
			if (line.startsWith("# APPOS")) {
				// format: # APPOS MAC X Y Z
				StringTokenizer st = new StringTokenizer(line);
				if ((apPositions != null) && (st.countTokens() == 6)) {
					if (st.nextToken().equals("#") && st.nextToken().equals("APPOS")) {
						MACAddress apMac = MACAddress.parse(st.nextToken());
						GeoPosition apPos = new GeoPosition(Double.parseDouble(st.nextToken()), Double.parseDouble(st.nextToken()), Double.parseDouble(st.nextToken()));
						apPositions.put(apMac, apPos);
					}
				}
				continue;
			}
			// skip ordinary comment
			if (line.startsWith("#"))
				continue;
			// split the line
			String[] components = line.split(";");
			for (String s: components) {
				String[] kv = s.split("=", 2);
				if (kv.length != 2)
					throw new IOException("Expected key=value");
				String key = kv[0];
				String value = kv[1];
			    // discard unnecessary parts
				if ((key.equals("t")) || (key.equals("id")))
					continue;
				// read the coordinates
				if ((key.equals("pos"))) {
					String[] coordinates = value.split(",");
					x = Double.parseDouble(coordinates[0]);
					y = Double.parseDouble(coordinates[1]);
					continue;
				}
				// read the direction
				if ((key.equals("degree"))) {
					degree = Double.parseDouble(value);
					degree = new Double(Math.round(degree / degmap) * degmap);
					if (degree == 360.0)
						degree = 0.0;
					continue;
				}
				// read received signal strength, frequency and mode
				String[] values = value.split(",", 4);
				if (values.length < 3)
					throw new IOException("Expected MAC=rss,frequency,mode,noise or MAC=rss,frequency,mode");
				double rss = Double.parseDouble(values[0]);
				int mode = Integer.parseInt(values[2]);
				// skip all non-infrastructure nodes
				if (mode != 3)
					continue;
				MACAddress mac = MACAddress.parse(key);
				// access point already known?
				if (macSignalStrengths.containsKey(mac)) { // ap known
					// get the map for this ap
					Map<GeoPosition, List<Double>> posSignalStrengthMeasurements = macSignalStrengths.get(mac);					
					// did the current position already occur?
					if (posSignalStrengthMeasurements.containsKey(new GeoPosition(x, y, 0.0, degree))) { // known position
						// get the list for this position
						List<Double> signalStrengthMeasurements = posSignalStrengthMeasurements.get(new GeoPosition(x, y, 0.0, degree));
						// add the reading
						signalStrengthMeasurements.add(rss);
					} else { // unknown position
						// create a new list of readings
						ArrayList<Double> signalStrengthMeasurements = new ArrayList<Double>();
						// insert the first reading
						signalStrengthMeasurements.add(rss);
						posSignalStrengthMeasurements.put(new GeoPosition(x, y, 0.0, degree), signalStrengthMeasurements);
					}
				} else { // ap unknown
					// create a new list of readings
					ArrayList<Double> signalStrengthMeasurements = new ArrayList<Double>();
					// insert the first reading
					if (rss == 0.0) System.out.println(rss);
					signalStrengthMeasurements.add(rss);
					// create a new map for this ap
					Map<GeoPosition, List<Double>> posSignalStrengthMeasurements = new HashMap<GeoPosition, List<Double>>();
					// add the list to the map
					posSignalStrengthMeasurements.put(new GeoPosition(x, y, 0.0, degree), signalStrengthMeasurements);
					// add the map to the global map
					macSignalStrengths.put(mac, posSignalStrengthMeasurements);
				}
				// add the direction to the list of directions
				if (!orientations.contains(degree))
					orientations.add(degree);
			}
		}
	}
	
	
	public static int analyzeRiceLog(BufferedReader file, Map<GeoPositionStep, Map<GeoPosition, Double>> probabilities, List<GeoPosition> realPositions, double[] minmax) throws IOException {
		int steps = 0;
		GeoPosition realPos = new GeoPosition();
		double min = 1.0;
		double max = 0.0;
		
		String line;
		while ((line = file.readLine()) != null) {
			if (line.startsWith("step=")) {
				Map<GeoPosition, Double> probabilityMap = new HashMap<GeoPosition, Double>();
				
				int step = Integer.parseInt(line.substring(line.indexOf("=") + 1, line.indexOf(";")));
				if ((step + 1) > steps)
					steps = (step + 1);
				String[] components = line.substring(line.indexOf(";") + 1).split(";");
				for (String s : components) {
					String[] subcomponents = s.split((","));
					GeoPosition pos = new GeoPosition(Double.parseDouble(subcomponents[0]), Double.parseDouble(subcomponents[1]));
					double value = Double.parseDouble(subcomponents[2]);
					probabilityMap.put(pos, value);
					// update global min and max values
					if (value < min)
						min = value;
					if (value > max)
						max = value;
				}
				probabilities.put(new GeoPositionStep(realPos, step), probabilityMap);
			} else if (line.startsWith("real=")) {
				realPos = new GeoPosition();
				String[] components = line.substring(line.indexOf("=") + 1).split(",");
				realPos.setX(Double.parseDouble(components[0]));
				realPos.setY(Double.parseDouble(components[1]));
				realPos.setZ(Double.parseDouble(components[2]));
				realPos.setOrientation(Double.parseDouble(components[3]));
				realPositions.add(realPos);
			}
		}
		minmax[0] = min;
		minmax[1] = max;
		return steps;
	}
	
}
